# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sublime
import sys
import os
import traceback
import time
import re

try:
    str_cls = unicode  # noqa
except (NameError):
    str_cls = str


__all__ = [
    'SvnInsertViewCommand',
    'SvnReplaceViewCommand',
    'SvnCommandFilterCommand',
    'SvnApplicableFileCommandsCommand',
    'SvnApplicableFolderCommandsCommand',
    'SvnApplicableWcCommandsCommand',
    'SvnFindingWorkingCopyCommand',
    'SvnNotInstalledCommand',
    'SvnCheckoutCommand',
    'SvnMenuCheckoutCommand',
    'SvnAddInteractiveCommand',
    'SvnBranchCommand',
    'SvnCleanupCommand',
    'SvnWcCommitCommand',
    'SvnCustomCommand',
    'SvnDeleteInteractiveCommand',
    'SvnWcDiffCommand',
    'SvnWcCustomDiffCommand',
    'SvnWcRemoteDiffCommand',
    'SvnRerunCommand',
    'SvnWcInfoCommand',
    'SvnWcLogCommand',
    'SvnMergeCommand',
    'SvnRelocateCommand',
    'SvnWcResolveTreeConflictCommand',
    'SvnRevertInteractiveCommand',
    'SvnWcStatusCommand',
    'SvnSwitchCommand',
    'SvnUnlockInteractiveCommand',
    'SvnWcUpdateCommand',
    'SvnWcUpdateIncludingExternalsCommand',
    'SvnUpgradeCommand',
    'SvnFolderAddCommand',
    'SvnFolderCopyCommand',
    'SvnFolderCommitCommand',
    'SvnFolderDeleteCommand',
    'SvnFolderDiffCommand',
    'SvnFolderCustomDiffCommand',
    'SvnFolderRemoteDiffCommand',
    'SvnFolderIgnoreCommand',
    'SvnFolderInfoCommand',
    'SvnFolderLogCommand',
    'SvnFolderMoveCommand',
    'SvnFolderPropertiesCommand',
    'SvnFolderResolvePropertyConflictCommand',
    'SvnFolderResolveTreeConflictCommand',
    'SvnFolderRevertCommand',
    'SvnFolderStatusCommand',
    'SvnFolderUnignoreCommand',
    'SvnFolderUpdateCommand',
    'SvnUpdateIncludingExternalsCommand',
    'SvnAddCommand',
    'SvnBlameCommand',
    'SvnCommitCommand',
    'SvnCopyCommand',
    'SvnDiffCommand',
    'SvnCustomDiffCommand',
    'SvnRemoteDiffCommand',
    'SvnDeleteCommand',
    'SvnFilePickCommand',
    'SvnFilePropertiesCommand',
    'SvnIgnoreCommand',
    'SvnInfoCommand',
    'SvnLockCommand',
    'SvnLogCommand',
    'SvnMoveCommand',
    'SvnResolveCommand',
    'SvnResolvePropertyConflictCommand',
    'SvnResolveTreeConflictCommand',
    'SvnRevertCommand',
    'SvnStatusCommand',
    'SvnUnignoreCommand',
    'SvnUnlockCommand',
    'SvnUpdateCommand',
    'SvnChangelistAddCommand',
    'SvnChangelistRemoveCommand',
    'SvnPropAddCommand',
    'SvnPropDeleteCommand',
    'SvnPropEditCommand',
    'SvnPropListCommand',
    'SvnMessageSaveListener',
    'SvnFocusListener',
    'SvnEnterLicenseKeyCommand',
    'SvnRemoveLicenseKeyCommand',
    'SvnEnableDebugModeCommand',
    'SvnDisableDebugModeCommand',
    'SvnEditSettingsCommand',
    'SvnOpenDefaultSettingsCommand',
    'SvnOpenUserSettingsCommand',
    'SvnEditKeyBindingsCommand',
    'SvnOpenDefaultKeyBindingsCommand',
    'SvnOpenUserKeyBindingsCommand',
]


st_version = 2
if int(sublime.version()) > 3000:
    st_version = 3

mod_reload_prefix = ''
if st_version == 3:
    mod_reload_prefix = 'SVN.'
    from imp import reload

globals_ = {}
if st_version == 3:
    globals_ = {'__name__': __name__}

need_package_control_upgrade = False


def do_import(module, names):
    return __import__(module, globals_, {}, names, 1)


reloading = {
    'happening': False,
    'shown': False
}

# We reload in reverse order to solve some issues with super()
reload_mods = []
for mod in sys.modules:
    if (mod[0:15] == 'SVN.sublimesvn.' or mod[0:11] == 'sublimesvn.' or mod == 'sublimesvn') and sys.modules[mod] is not None:
        reload_mods.append(mod)
        reloading['happening'] = True

if reloading['happening']:
    try:
        _temp_status = do_import('sublimesvn.status', ['StatusCache'])
        if hasattr(_temp_status.StatusCache, 'last_check_thread') and _temp_status.StatusCache.last_check_thread:
            _temp_status.StatusCache.last_check_thread.stop = True
        del _temp_status
    except (ImportError) as e:
        if str(e).find('bad magic number') != -1:
            need_package_control_upgrade = True
        else:
            raise

# Prevent popups during reload, saving the callbacks for re-adding later
if reload_mods:
    old_callbacks = {}
    hook_match = re.search(r"<class '(\w+).ExcepthookChain'>", str(sys.excepthook))
    if hook_match:
        _temp = __import__(hook_match.group(1), globals_, {}, ['ExcepthookChain'], -1)
        ExcepthookChain = _temp.ExcepthookChain
        old_callbacks = ExcepthookChain.names
    sys.excepthook = sys.__excepthook__

mods_load_order = [
    'sublimesvn',
    'sublimesvn.vendor.asn1crypto._errors',
    'sublimesvn.vendor.asn1crypto._int',
    'sublimesvn.vendor.asn1crypto._ordereddict',
    'sublimesvn.vendor.asn1crypto._teletex_codec',
    'sublimesvn.vendor.asn1crypto._types',
    'sublimesvn.vendor.asn1crypto._inet',
    'sublimesvn.vendor.asn1crypto._iri',
    'sublimesvn.vendor.asn1crypto.version',
    'sublimesvn.vendor.asn1crypto.pem',
    'sublimesvn.vendor.asn1crypto.util',
    'sublimesvn.vendor.asn1crypto.parser',
    'sublimesvn.vendor.asn1crypto.core',
    'sublimesvn.vendor.asn1crypto.algos',
    'sublimesvn.vendor.asn1crypto.keys',
    'sublimesvn.vendor.asn1crypto.x509',
    'sublimesvn.vendor.asn1crypto.crl',
    'sublimesvn.vendor.asn1crypto.csr',
    'sublimesvn.vendor.asn1crypto.ocsp',
    'sublimesvn.vendor.asn1crypto.cms',
    'sublimesvn.vendor.asn1crypto.pdf',
    'sublimesvn.vendor.asn1crypto.pkcs12',
    'sublimesvn.vendor.asn1crypto.tsp',
    'sublimesvn.vendor.asn1crypto',
    'sublimesvn.vendor.base58',
    'sublimesvn.ecdsa',
    'sublimesvn.win32',
    'sublimesvn.times',
    'sublimesvn.views',
    'sublimesvn.errors',
    'sublimesvn.cmd',
    'sublimesvn.debug',
    'sublimesvn.output',
    'sublimesvn.config',
    'sublimesvn.paths',
    'sublimesvn.threads',
    'sublimesvn.input',
    'sublimesvn.proc',
    'sublimesvn.status',
    'sublimesvn.commands',
    'sublimesvn.listeners',
    'sublimesvn.bloom_filter',
]

for mod in mods_load_order:
    if mod_reload_prefix + mod in reload_mods:
        reload(sys.modules[mod_reload_prefix + mod])

try:
    # Python 2
    from sublimesvn.commands import (
        SvnInsertViewCommand,
        SvnReplaceViewCommand,

        SvnCommandFilterCommand,
        SvnApplicableFileCommandsCommand,
        SvnApplicableFolderCommandsCommand,
        SvnApplicableWcCommandsCommand,
        SvnFindingWorkingCopyCommand,
        SvnNotInstalledCommand,

        SvnCheckoutCommand,
        SvnMenuCheckoutCommand,

        SvnAddInteractiveCommand,
        SvnBranchCommand,
        SvnCleanupCommand,
        SvnWcCommitCommand,
        SvnCustomCommand,
        SvnDeleteInteractiveCommand,
        SvnWcDiffCommand,
        SvnWcCustomDiffCommand,
        SvnWcRemoteDiffCommand,
        SvnRerunCommand,
        SvnWcInfoCommand,
        SvnWcLogCommand,
        SvnMergeCommand,
        SvnRelocateCommand,
        SvnWcResolveTreeConflictCommand,
        SvnRevertInteractiveCommand,
        SvnWcStatusCommand,
        SvnSwitchCommand,
        SvnUnlockInteractiveCommand,
        SvnWcUpdateCommand,
        SvnWcUpdateIncludingExternalsCommand,
        SvnUpgradeCommand,

        SvnFolderAddCommand,
        SvnFolderCopyCommand,
        SvnFolderCommitCommand,
        SvnFolderDeleteCommand,
        SvnFolderDiffCommand,
        SvnFolderCustomDiffCommand,
        SvnFolderRemoteDiffCommand,
        SvnFolderIgnoreCommand,
        SvnFolderInfoCommand,
        SvnFolderLogCommand,
        SvnFolderMoveCommand,
        SvnFolderPropertiesCommand,
        SvnFolderResolvePropertyConflictCommand,
        SvnFolderResolveTreeConflictCommand,
        SvnFolderRevertCommand,
        SvnFolderStatusCommand,
        SvnFolderUnignoreCommand,
        SvnFolderUpdateCommand,
        SvnUpdateIncludingExternalsCommand,

        SvnAddCommand,
        SvnBlameCommand,
        SvnCommitCommand,
        SvnCopyCommand,
        SvnDiffCommand,
        SvnCustomDiffCommand,
        SvnRemoteDiffCommand,
        SvnDeleteCommand,
        SvnFilePickCommand,
        SvnFilePropertiesCommand,
        SvnIgnoreCommand,
        SvnInfoCommand,
        SvnLockCommand,
        SvnLogCommand,
        SvnMoveCommand,
        SvnResolveCommand,
        SvnResolvePropertyConflictCommand,
        SvnResolveTreeConflictCommand,
        SvnRevertCommand,
        SvnStatusCommand,
        SvnUnignoreCommand,
        SvnUnlockCommand,
        SvnUpdateCommand,

        SvnChangelistAddCommand,
        SvnChangelistRemoveCommand,

        SvnPropAddCommand,
        SvnPropDeleteCommand,
        SvnPropEditCommand,
        SvnPropListCommand,

        SvnEnterLicenseKeyCommand,
        SvnRemoveLicenseKeyCommand,

        SvnEnableDebugModeCommand,
        SvnDisableDebugModeCommand,

        SvnEditSettingsCommand,
        SvnOpenDefaultSettingsCommand,
        SvnOpenUserSettingsCommand,

        SvnEditKeyBindingsCommand,
        SvnOpenDefaultKeyBindingsCommand,
        SvnOpenUserKeyBindingsCommand,

        cmd_init,
        cmd_cleanup
    )

    from sublimesvn.listeners import SvnMessageSaveListener, SvnFocusListener

    from sublimesvn.status import StatusCache
    from sublimesvn.proc import SVN
    from sublimesvn.errors import NotFoundError
    from sublimesvn.config import init_config

    from sublimesvn import (
        debug as svn_debug,
        times as svn_times,
        paths as svn_paths,
        threads as svn_threads
    )
except (ImportError):
    try:
        # Python 3
        from .sublimesvn.commands import (
            SvnInsertViewCommand,
            SvnReplaceViewCommand,

            SvnCommandFilterCommand,
            SvnApplicableFileCommandsCommand,
            SvnApplicableFolderCommandsCommand,
            SvnApplicableWcCommandsCommand,
            SvnFindingWorkingCopyCommand,
            SvnNotInstalledCommand,

            SvnCheckoutCommand,
            SvnMenuCheckoutCommand,

            SvnAddInteractiveCommand,
            SvnBranchCommand,
            SvnCleanupCommand,
            SvnWcCommitCommand,
            SvnCustomCommand,
            SvnDeleteInteractiveCommand,
            SvnWcDiffCommand,
            SvnWcCustomDiffCommand,
            SvnWcRemoteDiffCommand,
            SvnRerunCommand,
            SvnWcInfoCommand,
            SvnWcLogCommand,
            SvnMergeCommand,
            SvnRelocateCommand,
            SvnWcResolveTreeConflictCommand,
            SvnRevertInteractiveCommand,
            SvnWcStatusCommand,
            SvnSwitchCommand,
            SvnUnlockInteractiveCommand,
            SvnWcUpdateCommand,
            SvnWcUpdateIncludingExternalsCommand,
            SvnUpgradeCommand,

            SvnFolderAddCommand,
            SvnFolderCopyCommand,
            SvnFolderCommitCommand,
            SvnFolderDeleteCommand,
            SvnFolderDiffCommand,
            SvnFolderCustomDiffCommand,
            SvnFolderRemoteDiffCommand,
            SvnFolderIgnoreCommand,
            SvnFolderInfoCommand,
            SvnFolderLogCommand,
            SvnFolderMoveCommand,
            SvnFolderPropertiesCommand,
            SvnFolderResolvePropertyConflictCommand,
            SvnFolderResolveTreeConflictCommand,
            SvnFolderRevertCommand,
            SvnFolderStatusCommand,
            SvnFolderUnignoreCommand,
            SvnFolderUpdateCommand,
            SvnUpdateIncludingExternalsCommand,

            SvnAddCommand,
            SvnBlameCommand,
            SvnCommitCommand,
            SvnCopyCommand,
            SvnDiffCommand,
            SvnCustomDiffCommand,
            SvnRemoteDiffCommand,
            SvnDeleteCommand,
            SvnFilePickCommand,
            SvnFilePropertiesCommand,
            SvnIgnoreCommand,
            SvnInfoCommand,
            SvnLockCommand,
            SvnLogCommand,
            SvnMoveCommand,
            SvnResolveCommand,
            SvnResolvePropertyConflictCommand,
            SvnResolveTreeConflictCommand,
            SvnRevertCommand,
            SvnStatusCommand,
            SvnUnignoreCommand,
            SvnUnlockCommand,
            SvnUpdateCommand,

            SvnChangelistAddCommand,
            SvnChangelistRemoveCommand,

            SvnPropAddCommand,
            SvnPropDeleteCommand,
            SvnPropEditCommand,
            SvnPropListCommand,

            SvnEnterLicenseKeyCommand,
            SvnRemoveLicenseKeyCommand,

            SvnEnableDebugModeCommand,
            SvnDisableDebugModeCommand,

            SvnEditSettingsCommand,
            SvnOpenDefaultSettingsCommand,
            SvnOpenUserSettingsCommand,

            SvnEditKeyBindingsCommand,
            SvnOpenDefaultKeyBindingsCommand,
            SvnOpenUserKeyBindingsCommand,

            cmd_init,
            cmd_cleanup
        )

        from .sublimesvn.listeners import SvnMessageSaveListener, SvnFocusListener

        from .sublimesvn.status import StatusCache
        from .sublimesvn.proc import SVN
        from .sublimesvn.errors import NotFoundError
        from .sublimesvn.config import init_config

        from .sublimesvn import (
            debug as svn_debug,
            times as svn_times,
            paths as svn_paths,
            threads as svn_threads
        )
    except (ImportError) as e:
        if str(e).find('bad magic number') != -1:
            need_package_control_upgrade = True
        else:
            raise

hook_match = re.search(r"<class '(\w+).ExcepthookChain'>", str(sys.excepthook))

if not hook_match:
    class ExcepthookChain(object):
        callbacks = []
        names = {}

        @classmethod
        def add(cls, name, callback):
            if name == 'sys.excepthook':
                if name in cls.names:
                    return
                cls.callbacks.append(callback)
            else:
                if name in cls.names:
                    cls.callbacks.remove(cls.names[name])
                cls.callbacks.insert(0, callback)
            cls.names[name] = callback

        @classmethod
        def hook(cls, type, value, tb):
            for callback in cls.callbacks:
                callback(type, value, tb)

        @classmethod
        def remove(cls, name):
            if name not in cls.names:
                return
            callback = cls.names[name]
            del cls.names[name]
            cls.callbacks.remove(callback)
else:
    _temp = __import__(
        hook_match.group(1),
        globals(),
        locals(),
        ['ExcepthookChain'],
        -1
    )
    ExcepthookChain = _temp.ExcepthookChain


# Override default uncaught exception handler
def svn_uncaught_except(type, value, tb):
    message = ''.join(traceback.format_exception(type, value, tb))

    if message.find('/sublimesvn/') != -1 or message.find('\\sublimesvn\\') != -1:
        def append_log():
            log_file_path = os.path.join(sublime.packages_path(), 'User', 'SVN.errors.log')
            send_log_path = log_file_path
            timestamp = svn_times.timestamp_to_string(time.time(), '%Y-%m-%d %H:%M:%S\n')
            with open(log_file_path, 'a') as f:
                f.write(timestamp)
                f.write(message)
            if svn_debug.get_debug() and svn_debug.get_debug_log_file():
                send_log_path = svn_debug.get_debug_log_file()
                svn_debug.debug_print(message)
            sublime.error_message(
                'SVN\n\nAn unexpected error occurred, please submit '
                'the file %s with a Request at https://codexns.io/account' % send_log_path
            )
            sublime.active_window().run_command(
                'open_file',
                {'file': svn_paths.fix_windows_path(send_log_path)}
            )
        if reloading['happening']:
            if not reloading['shown']:
                sublime.error_message(
                    'SVN\n\nSVN was just upgraded, please restart Sublime to finish the upgrade'
                )
                reloading['shown'] = True
        else:
            sublime.set_timeout(append_log, 10)


if reload_mods and old_callbacks:
    for name in old_callbacks:
        ExcepthookChain.add(name, old_callbacks[name])

ExcepthookChain.add('sys.excepthook', sys.__excepthook__)
ExcepthookChain.add('svn_uncaught_except', svn_uncaught_except)

if sys.excepthook != ExcepthookChain.hook:
    sys.excepthook = ExcepthookChain.hook


settings_dict = {'svn_binary_path': None}


try:
    # This won't work if the import failed
    class SvnInit(svn_threads.HookedThread):
        def run_with_except(self):
            try:
                SVN.init(svn_binary_path=settings_dict['svn_binary_path'])
            except (NotFoundError):
                def show_not_found():
                    if SVN.shown_global_not_found:
                        return
                    SVN.shown_global_not_found = True
                    SVN.version = False

                    message = 'SVN\n\nThe Subversion command line client could not be found on your machine.'
                    if int(sublime.version()) >= 2190:
                        message += '\n\nWould you like to visit the SVN site for installation help?'
                        if sublime.ok_cancel_dialog(message, 'Yes'):
                            sublime.active_window().run_command(
                                'open_url',
                                {'url': 'https://codexns.io/products/svn_for_sublime/install#Subversion_Command_Line'}
                            )
                        else:
                            return
                    else:
                        message += ' Please visit https://codexns.io/svn for help.'
                        sublime.error_message(message)
                        return
                sublime.set_timeout(show_not_found, 1)

            except (OSError) as e:
                svn_path = re.sub(r'The binary "(.*)" does not exist.*$', '\\1', str_cls(e))
                error = str_cls(e)
                if svn_path not in SVN.shown_missing:
                    def do_show():
                        sublime.error_message('SVN\n\n' + error)
                    sublime.set_timeout(do_show, 1)
                    SVN.shown_missing[svn_path] = True
except (NameError):
    pass


def plugin_loaded():
    if need_package_control_upgrade:
        sublime.error_message(
            'SVN\n\nThe SVN package seems to have been '
            'installed using an older version of Package Control. Please '
            'remove the SVN package, upgrade Package Control to 2.0.0 '
            'and then reinstall SVN.\n\nIt may be necessary to delete '
            'the "Packages/Package Control/" folder and then follow the '
            'instructions at https://packagecontrol.io/installation to '
            'properly upgrade Package Control.'
        )
        return

    init_config()
    settings = sublime.load_settings('SVN.sublime-settings')
    svn_debug.set_debug(settings.get('debug', False))
    svn_debug.set_debug_log_file(settings.get('debug_log_file', None))

    def reset_shown():
        SVN.shown_missing = {}

    cmd_init()

    settings.add_on_change('svn_binary_path', reset_shown)
    settings.add_on_change('auto_update_check_frequency', StatusCache.set_check_updates)
    StatusCache.set_check_updates()

    settings_dict['svn_binary_path'] = settings.get('svn_binary_path')

    SvnInit().start()

    bin_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin')
    has_bin = os.path.exists(bin_folder)
    svn_exe = os.path.join(bin_folder, '1.8', 'svn.exe')
    has_svn = os.path.exists(svn_exe)
    if os.name == 'nt' and (not has_bin or not has_svn):
        sublime.error_message(
            'SVN\n\nThe SVN package seems to have been '
            'synced or copied from an OS X or Linux machine. The Windows '
            'version of the package is different due to the inclusion of '
            'a number of necessary exe files.\n\nTo fix the SVN package '
            'so that it may run properly, please run "Remove Package" and '
            'then reinstall it using the "Install Package" command.\n\nTo '
            'learn how to properly sync packages across different machines, '
            'please visit https://packagecontrol.io/docs/syncing'
        )


if st_version == 2:
    plugin_loaded()


def plugin_unloaded():
    cmd_cleanup()
    ExcepthookChain.remove('svn_uncaught_except')


if st_version == 2:
    def unload_handler():
        plugin_unloaded()
